import os
import json
import statistics

import pandas as pd
import numpy as np


class DataSet:
    def __init__(self, D, name="default"):
        self.D = D
        self.n, self.d = self.D.shape
        self.name = name

    def getResource(self, index):
        return self.D.iloc[index, :]

    def saveMetaData(self, filepath):
        metadata = dict(d=self.d, n=self.n, name=self.name)
        string = json.dumps(metadata, indent=2)
        with open(filepath, 'w') as f:
            f.write(string)
        return string

    def getMatrix(self):
        return self.D.as_matrix()


def convertDtype(l):
    try:
        return np.array(l, dtype="float")
    except:
        pass
    l = np.array(l, dtype=str)
    l[l == 'nan'] = 'NA'
    return l


class CSVDataSet(DataSet):
    """ A dataset living locally in a .csv file

    """

    def __init__(self,
                 csv_path,
                 index_column=None,
                 NA_val=".",
                 name="mydataset"):
        self.name = name

        # Load the data set
        D = pd.read_csv(csv_path, dtype="unicode")
        self.n, self.d = D.shape
        print("Dataset of size", self.n, "samples", self.d, "dimensions",
              "Loaded")

        # Convert to numeric all numeric rows
        D = D.replace(NA_val, "nan")
        print("Replacing all", NA_val, "with nan")
        d = []
        for c in D.columns:
            d.append(convertDtype(list(D[c])))
            print("Converting", c, end="\r\r")
        newcolumns = D.columns
        newindex = D.index
        D = list(d)
        D = pd.DataFrame(dict(zip(newcolumns, D)), index=newindex)

        # Set the index column as specified
        if index_column is not None:
            print("Setting index column as", index_column)
            D.index = D[index_column]
            print("Deleting", index_column, "from dataset")
            del D[index_column]

        self.D = D

        # Remove all columns which have all null values
        keep = []
        allnull = self.D.isnull().all(axis=0)
        for c in self.D.columns[allnull]:
            print("Removing column", c, "because it has all null values")
        keep = self.D.columns[~allnull]
        self.D = self.D[keep]

        # Remove all rows which have all null values
        allnull = self.D.isnull().all(axis=1)
        for r in self.D.index[allnull]:
            print("Removing row", r, "because it has all null values")
        keep = self.D.index[~allnull]
        self.D = self.D.loc[keep]
        n, d = self.D.shape
        print("Dataset of size", n, "samples", d, "dimensions", "Resulting")
        self.N = self.D.shape[0]

    def imputeColumns(self, numeric):
        keep = []
        keep = (self.D.dtypes == "float64").as_matrix()
        for c in self.D.columns[~keep]:
            print("Removing column", c, "because it is not numeric")
        self.D = self.D[self.D.columns[keep]]
        cmean = self.D.mean(axis=0)
        values = dict(list(zip(self.D.columns, cmean.as_matrix())))
        #self.D.fillna(value=values, inplace=True)
        d = self.D.as_matrix()
        for i, c in enumerate(self.D.columns):
            print("Imputing column", c, "with value", values[c])
            d[:, i][np.isnan(d[:, i])] = values[c]
        D = pd.DataFrame(d)
        D.index = self.D.index
        D.index.names = self.D.index.names
        D.columns = self.D.columns
        D.columns.names = self.D.columns.names
        self.D = D
        allzero = np.all(self.D.as_matrix() == 0, axis=0)
        for c in self.D.columns[allzero]:
            print("Removing column", c, "because it has all zero values")
        keep = self.D.columns[~allzero]
        allsame = np.std(self.D.as_matrix(), axis=0) == 0
        for c in self.D.columns[allsame]:
            print(
                "Removing column", c,
                "because it has all zero standard deviation (all values same)")
        keep = self.D.columns[~allsame]
        self.D = self.D[keep]
        n, d = self.D.shape
        print("Dataset of size", n, "samples", d, "dimensions", "Resulting")
        print("Dataset has", self.D.isnull().sum().sum(), "nans")
        print("Dataset has", np.sum(np.isinf(self.D.as_matrix())), "infs")

    def getResource(self, index):
        """Get a specific data point from the data set.

        Parameters
        ----------
        index : int or string
            The index of the data point in `D`, either positional or a string.

        Returns
        -------
        :obj:`ndarray`
            A ndarray of the data point.

        """
        if type(index) is int:
            return self.D.iloc[index].as_matrix()
        else:
            return self.D.loc[index].as_matrix()

    def getColumn(self, index):
        """Get a column of the dataframe.
 
        Parameters
        ----------
        index : int or string
            The index of the column in `D`, either positional or a string.

        Returns
        -------
        :obj:`ndarray`
            The values in the column.
        """
        if type(index) is int:
            return self.D.iloc[:, index].as_matrix()
        else:
            return self.D[index].as_matrix()

    def getColumnValues(self, index):
        """Get the unique values of a column.

        Parameters
        ----------
        index : int or string
            The index of the column in `D`, either positional or a string.

        Returns
        -------
        :obj:`ndarray`
            A ndarray of the unique values.

        """
        column = self.getColumn(index)
        if column.dtype == "float64":
            column = column[~np.isnan(column)]
        else:
            column = column[np.array([x != "NA" for x in column])]
        return np.unique(column)

    def getColumnDistribution(self, index):
        """Get the distribution of values in a column.

        Parameters
        ----------
        index : int or string
            The index of the column in `D`, either positional or a string.

        Returns
        -------
        :obj:`ndarray`, :obj:`ndarray`
            An array x of the unique labels, and an array y of the count of that label

        """
        x = self.getColumnValues(index)
        column = self.getColumn(index)
        y = [np.sum(column == v) for v in x]
        return x, y

    def getColumnNADist(self, index):
        column = self.getColumn(index)
        if column.dtype == "float64":
            na = np.sum([np.isnan(x) for x in column])
            not_na = len(column) - na
            return na, not_na
        else:
            na = np.sum([x == "NA" for x in column])
            not_na = len(column) - na
            return na, not_na
        return na, not_na

    def getColumnDescription(self, index, sep="\n"):
        """Get a description of the column.

        """
        desc = []
        if type(index) is int:
            index = self.D.columns.values[index]
        for i, name in enumerate(self.D.columns.names):
            desc.append(name + ": " + index[i])
        return sep.join(desc)

    def getLevelValues(self, index):
        return np.unique(self.D.columns.get_level_values(index))



import pkg_resources
import time
from collections import OrderedDict

import numpy as np
import pandas as pd
import knor
from scipy.stats import zscore
from plotly.offline import (download_plotlyjs, init_notebook_mode)
from jinja2 import Environment, PackageLoader, select_autoescape, FileSystemLoader

import pymeda.datasets as lds
import pymeda.plotters as lpl
import pymeda.clustering as lcl

np.warnings.filterwarnings('ignore')


class Meda:
    """ 
    A meda object contains the data that will be plotted through the
    class functions. 

    Parameters
    ----------
    data : str or pd.DataFrame object
        String that specifies path to a csv file or 
        pd.DataFrame object.
    title : str
        Title of the dataset.
    index_col : str, optional
        Name of the index column in csv file if `data` is a string.
    cluster_levels : int (default=1), or None
        Number of levels for hierarchial clustering (e.g. 1 means running hierarchial
        clustering only once. 2 means run twice). If None, it does not run clustering.
    showticklabels : bool, optional
        If default value is passed (True), all the dimension labels will be displayed.
        Set to False to hide all dimension labels. 
    mode : str, optional
        String that determines if plotting in a Jupyter notebook or for static
        HTML embedding. Defaults to 'notebook'. Set to 'div' for HTML embedding.
    """

    def __init__(self,
                 data,
                 title,
                 index_col=None,
                 cluster_levels=1,
                 showticklabels=True,
                 mode='notebook'):
        #Check if input is a path to csv file or pd.DataFrame object
        if isinstance(data, str):
            self._ds = lds.CSVDataSet(data, name=title, index_column=index_col)
        elif isinstance(data, pd.DataFrame):
            self._ds = lds.DataSet(data, name=title)
        else:
            raise TypeError

        self._ds_normed = lds.DataSet(self._ds.D.apply(zscore), name=title)
        self._showticklabels = showticklabels
        self._cluster_levels = cluster_levels
        self._mode = mode
        if mode == 'notebook':
            init_notebook_mode()
        self._cluster_ds = None

    def _compute_clusters(self, cluster_mode=None):
        """
        Run hierarchial GMM clustering

        Parameters
        ----------
        cluster_mode : str 
            Not implemented yet
        """
        cluster_ds = lcl.HGMMClustering(
            self._ds_normed, levels=self._cluster_levels)
        cluster_ds.cluster()
        self._cluster_ds = cluster_ds

    def ridgeline(self, mode=None):
        """
        Generate ridgeline plots of all observations.
        """
        if not mode:
            mode = self._mode

        return lpl.RidgeLine(self._ds_normed, mode=mode).plot()

    def heatmap(self, mode=None):
        """
        Generate heatmap of all observations
        If n > 1000, then a kmeans++ initializaiton is performed to derive 
        1000 cluster centers. 
        """
        if not mode:
            mode = self._mode

        if self._ds.n > 1000:  #if sample size is > 1000, run kmeans++ initialization
            ret = knor.Kmeans(
                self._ds_normed.D.values, 1000, max_iters=0, init='kmeanspp')
            centroids_df = pd.DataFrame(
                ret.get_centroids(), columns=self._ds.D.columns)
            centroids_ds = lds.DataSet(centroids_df, name=self._ds.name)

            return lpl.Heatmap(
                centroids_ds, mode=mode).plot(showticklabels=True)
        else:
            return lpl.Heatmap(
                self._ds_normed, mode=mode).plot(showticklabels=True)

    def histogram_heatmap(self, mode=None):
        """
        Generate 1d heatmap
        """
        if not mode:
            mode = self._mode

        if self._ds.n > 1000:  #if sample size is > 1000, run kmeans++ initialization
            ret = knor.Kmeans(
                self._ds_normed.D.values, 1000, max_iters=0, init='kmeanspp')
            centroids_df = pd.DataFrame(
                ret.get_centroids(), columns=self._ds.D.columns)
            centroids_ds = lds.DataSet(centroids_df, name=self._ds.name)

            return lpl.HistogramHeatmap(
                centroids_ds,
                mode=mode).plot(showticklabels=self._showticklabels)
        else:
            return lpl.HistogramHeatmap(
                self._ds_normed,
                mode=mode).plot(showticklabels=self._showticklabels)

    def scree_plot(self, mode=None):
        """
        Genereate scree plot of PCA results
        """
        if not mode:
            mode = self._mode

        return lpl.ScreePlotter(self._ds_normed, mode=mode).plot()

    def correlation_matrix(self, mode=None):
        """
        Correlation matrix 
        """
        if not mode:
            mode = self._mode

        return lpl.CorrelationMatrix(
            self._ds_normed,
            mode=mode).plot(showticklabels=self._showticklabels)

    def cluster_dendrogram(self, mode=None):
        """
        Generate GMM dendrogram
        """
        if not mode:
            mode = self._mode

        if not self._cluster_ds:
            self._compute_clusters()

        return lpl.HierarchicalClusterMeansDendrogram(
            self._cluster_ds, mode=mode).plot()

    def cluster_means_stacked(self, mode=None):
        """
        Generate hierarchial clustering stacked means plot
        """
        if not mode:
            mode = self._mode

        if not self._cluster_ds:
            self._compute_clusters()

        return lpl.HierarchicalStackedClusterMeansHeatmap(
            self._cluster_ds,
            mode=mode).plot(showticklabels=self._showticklabels)

    def cluster_means_heatmap(self, mode=None):
        """
        Generate hierarchial cluster means plot of the lowest level
        """
        if not mode:
            mode = self._mode

        if not self._cluster_ds:
            self._compute_clusters()

        return lpl.ClusterMeansLevelHeatmap(
            self._cluster_ds,
            mode=mode).plot(showticklabels=self._showticklabels)

    def cluster_means_lines(self, mode=None):
        """
        Generate hierarchial cluster mean lines plot of the lowest level
        """
        if not mode:
            mode = self._mode

        if not self._cluster_ds:
            self._compute_clusters()

        return lpl.ClusterMeansLevelLines(
            self._cluster_ds,
            mode=mode).plot(showticklabels=self._showticklabels)

    def cluster_pair_plot(self, mode=None):
        """
        Generate pair plot for clusters
        """
        if not mode:
            mode = self._mode

        if not self._cluster_ds:
            self._compute_clusters()

        return lpl.ClusterPairsPlot(self._cluster_ds, mode=mode).plot()

    def location_lines(self, mode=None):
        """
        Generate location estimates using line graph
        """
        if not mode:
            mode = self._mode

        return lpl.LocationLines(
            self._ds, mode=mode).plot(showticklabels=self._showticklabels)

    def location_heatmap(self, mode=None):
        """
        Generate location estimates using heatmap
        """
        if not mode:
            mode = self._mode

        return lpl.LocationHeatmap(
            self._ds, mode=mode).plot(showticklabels=self._showticklabels)

    def run_all(self, mode=None):
        """
        Run all plots available in PyMEDA.

        Parameters
        ----------
        mode : str, optional 
        """
        if not mode:
            mode = self._mode

        #Make plots and save as div
        heatmap = self.heatmap(mode=mode)
        ridgeline = self.ridgeline(mode=mode)
        location_heatmap = self.location_heatmap(mode=mode)
        location_lines = self.location_lines(mode=mode)

        scree_plot = self.scree_plot(mode=mode)
        corr_matrix = self.correlation_matrix(mode=mode)

        #HGMM plots
        if self._cluster_levels is not None:
            hgmm_dendogram = self.cluster_dendrogram(mode=mode)
            if not self._ds.n > 1000:
                hgmm_pair_plot = self.cluster_pair_plot(mode=mode)
            hgmm_stacked_mean = self.cluster_means_stacked(mode=mode)
            hgmm_cluster_mean = self.cluster_means_heatmap(mode=mode)
            hgmm_cluster_means = self.cluster_means_lines(mode=mode)

        if mode == 'div':
            if not self._ds.n > 1000:
                titles = [
                    "Representative Heatmap", "Ridge Line Plot",
                    "Location Heatmap", "Location Lines", "Correlation Matrix",
                    "Scree Plot", "Hierarchical GMM Dendogram", "Pair Plot",
                    "Cluster Stacked Means", "Cluster Mean Heatmap",
                    "Cluster Mean Lines"
                ]

                data = [
                    heatmap, ridgeline, location_heatmap, location_lines,
                    corr_matrix, scree_plot, hgmm_dendogram, hgmm_pair_plot,
                    hgmm_stacked_mean, hgmm_cluster_mean, hgmm_cluster_means
                ]

                out = OrderedDict(zip(titles, data))
            else:
                titles = [
                    "Representative Heatmap", "Ridge Line Plot",
                    "Location Heatmap", "Location Lines", "Correlation Matrix",
                    "Scree Plot", "Hierarchical GMM Dendogram",
                    "Cluster Stacked Means", "Cluster Mean Heatmap",
                    "Cluster Mean Lines"
                ]

                data = [
                    heatmap, ridgeline, location_heatmap, location_lines,
                    corr_matrix, scree_plot, hgmm_dendogram, hgmm_stacked_mean,
                    hgmm_cluster_mean, hgmm_cluster_means
                ]

                out = OrderedDict(zip(titles, data))
            return out

    def generate_report(self, out):
        """
        Creates a report

        Parameters
        ----------
        out : str
            Path to output the report
        """
        #Generate plots
        mode = 'div'
        plots = self.run_all(mode=mode)

        #Variables for output report name
        date = time.strftime("%Y-%m-%d")
        title = self._ds.name

        #Using package directory
        path = '/templates/'
        template_path = pkg_resources.resource_filename(__name__, path)

        env = Environment(loader=FileSystemLoader([template_path]))
        template = env.get_template('report.html')
        report = template.render(title=title, plots=plots, date=date)

        if out.endswith('/'):
            out = out + date + '_' + title + '.html'
        else:
            out = out + '/' + date + '_' + title + '.html'

        with open(out, 'w') as f:
            f.write(report)

        print('Report saved at {}'.format(out))


import os

from plotly.offline import iplot, plot
import plotly.graph_objs as go
import plotly.figure_factory as ff
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import scipy.stats as stats
import colorlover as cl


def get_spaced_colors(n):
    max_value = 255
    interval = int(max_value / n)
    hues = range(0, max_value, interval)
    return cl.to_rgb(["hsl(%d,100%%,40%%)" % i for i in hues])


def get_heat_colors(n):
    max_value = 255
    interval = int(max_value / n)
    hues = range(0, max_value, interval)
    return cl.to_rgb(["hsl(%d,100%%,40%%)" % i for i in hues])


def get_plt_cmap(cmap, n):
    """
    Helper function that converts matplotlib cmap to 
    integers in R, G, B space.

    Parameters
    ----------
    cmap : str
        Colormap from matplotlib 
    n : int
        Number of colors to output
        
    Returns
    -------
    out : list
        List of RGB values in format that can be used in Plotly
    """
    ranges = np.linspace(0, 1, num=n)
    arr = plt.cm.get_cmap(cmap)(ranges)
    arr = arr[:, :3] * 255

    out = []
    for r, g, b in arr.astype(np.int):
        out.append('rgb({},{},{})'.format(r, g, b))

    return out


class MatrixPlotter:
    def __init__(self, DS, mode="notebook", base_path=None):
        self.DS = DS
        self.plot_mode = mode
        self.base_path = base_path

        Reds = cl.scales['8']['seq']['Reds']
        self.Reds = list(zip(np.linspace(0, 1, len(Reds)), Reds))

        BuRd = cl.scales['11']['div']['RdBu'][::-1]
        self.BuRd = list(zip(np.linspace(0, 1, len(BuRd)), BuRd))

    def makeplot(self, fig, local_path=None):
        """Make the plotly figure visable to the user in the way they want.

        Parameters
        ----------
        gid : :obj:`figure`
            An plotly figure.

        """
        if self.plot_mode == "notebook":
            iplot(fig)
        if self.plot_mode == "savediv":
            fig["layout"]["autosize"] = True
            div = plot(fig, output_type='div', include_plotlyjs=False)
            path = os.path.join(self.base_path, local_path + ".html")
            os.makedirs("/".join(path.split("/")[:-1]), exist_ok=True)
            with open(path, "w") as f:
                f.write(div)
                f.close()

        if self.plot_mode == "div":
            fig["layout"]["autosize"] = True
            return plot(fig, output_type='div', include_plotlyjs=False)

    def _get_layout(self, title, xaxis, yaxis):
        if self.plot_mode == "div":
            return dict(xaxis=xaxis, yaxis=yaxis)
        else:
            return dict(title=title, xaxis=xaxis, yaxis=yaxis)


class Heatmap(MatrixPlotter):
    titlestring = "%s Heatmap"
    shortname = "heatmap"

    def plot(self, showticklabels=False):
        title = self.titlestring % (self.DS.name)
        xaxis = go.XAxis(
            title="Observations",
            ticktext=self.DS.D.index,
            ticks="",
            showticklabels=False,
            tickvals=[i for i in range(len(self.DS.D.index))])
        yaxis = go.YAxis(
            title="Dimensions",
            ticktext=self.DS.D.columns,
            ticks="",
            showticklabels=showticklabels,
            tickvals=[i for i in range(len(self.DS.D.columns))])
        layout = self._get_layout(title, xaxis, yaxis)

        maximum = self.DS.D.max().max()
        trace = go.Heatmap(
            z=self.DS.D.as_matrix().T,
            zmin=-maximum,
            zmax=maximum,
            colorscale=self.BuRd)
        data = [trace]
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)


class LocationHeatmap(MatrixPlotter):
    titlestring = "%s Location Heatmap"
    shortname = "locationheat"

    def plot(self, showticklabels=False):
        title = self.titlestring % (self.DS.name)
        D = self.DS.D.as_matrix().T
        means = np.mean(D, axis=1)
        medians = np.median(D, axis=1)
        z = np.vstack([means, medians])
        yaxis = go.YAxis(
            ticktext=["mean", "median"], showticklabels=True, tickvals=[0, 1])
        xaxis = go.XAxis(title="dimensions", showticklabels=showticklabels)
        layout = self._get_layout(title, xaxis, yaxis)

        trace = go.Heatmap(x=self.DS.D.columns, z=z, colorscale=self.Reds)
        data = [trace]
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)


class LocationLines(MatrixPlotter):
    titlestring = "%s Embedding Location Lines"
    shortname = "locationlines"

    def plot(self, showticklabels=False):
        title = self.titlestring % (self.DS.name)
        D = self.DS.D.as_matrix().T
        means = np.mean(D, axis=1)
        medians = np.median(D, axis=1)
        trace0 = go.Scatter(x=self.DS.D.columns, y=means, name="means")
        trace1 = go.Scatter(x=self.DS.D.columns, y=medians, name="medians")

        xaxis = dict(title="Dimensions", showticklabels=showticklabels)
        yaxis = dict(title="Mean or Median Value")
        layout = self._get_layout(title, xaxis, yaxis)

        data = [trace0, trace1]
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)


class HistogramHeatmap(MatrixPlotter):
    titlestring = "%s Histogram Heatmap"
    shortname = "histogramheat"

    def plot(self, showticklabels=False, scale=None):
        title = self.titlestring % (self.DS.name)
        D = self.DS.D.as_matrix().T
        d, n = D.shape
        D = (D - np.mean(D, axis=1).reshape(d, 1)) / np.std(
            D, axis=1).reshape(d, 1)
        D = np.nan_to_num(D)  # only nan if std all 0 -> all values 0

        num_bins = int(np.sqrt(2 * n))
        if num_bins > 20:
            num_bins = 20
        min_val = np.floor(np.min(D))
        if min_val < -5:
            min_val = -5
        max_val = np.ceil(np.max(D))
        if max_val > 5:
            max_val = 5
        bins = np.linspace(min_val, max_val,
                           (max_val - min_val) * num_bins + 1)
        bin_centers = (bins[1:] + bins[:-1]) / 2
        H = []
        for i in range(D.shape[0]):
            hist = np.histogram(D[i, :], bins=bins)[0]
            H.append(hist)
        z = np.vstack(H).astype(np.float)

        if scale == 'log':
            z[z > 0] = np.log(z[z > 0], dtype=np.float)

        trace = go.Heatmap(
            y=self.DS.D.columns,
            z=z,
            x=bins,
            colorscale=self.Reds,
            colorbar=go.ColorBar(title='Counts'))
        data = [trace]
        xaxis = go.XAxis(
            title="Normalized Value",
            ticks="outside",
            showticklabels=True,
        )
        yaxis = go.YAxis(
            title="Dimensions",
            ticks="",
            showticklabels=showticklabels,
            mirror=True)
        layout = self._get_layout(title, xaxis, yaxis)
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)


class RidgeLine(MatrixPlotter):
    titlestring = "%s Ridgeline Plot"
    shortname = "ridgeline"

    def plot(self):
        title = self.titlestring % (self.DS.name)
        D = self.DS.D.as_matrix().T
        columns = self.DS.D.columns[::-1]
        d, n = D.shape
        # Standardize each feature so that mean=0, std=1
        D = (D - np.mean(D, axis=1).reshape(d, 1)) / np.std(
            D, axis=1).reshape(d, 1)
        D = np.nan_to_num(D)  # only nan if std all 0 -> all values 0

        # Get colors
        colors = get_plt_cmap('rainbow', d)

        # Clip the min and max values at -5 and 5 respectively
        min_val = np.floor(np.min(D))
        if min_val < -5:
            min_val = -5
        max_val = np.ceil(np.max(D))
        if max_val > 5:
            max_val = 5

        x_range = np.linspace(min_val, max_val, 100)

        # calculate guassian KDEs
        kdes = []
        for row in D:
            kde = stats.kde.gaussian_kde(row)
            kdes.append(kde(x_range))

        # Spacing between each ridgeline
        spacing = 0.5

        # Plot each ridgelines
        data = []
        for idx, y in enumerate(kdes[::-1]):
            y += idx * spacing  # Amount to separate each ridgeline
            trace = go.Scatter(
                x=x_range,
                y=y,
                name=columns[idx],
                mode='lines',
                line=dict(color='rgb(0,0,0)', width=1.5),
                fill='toself',
                fillcolor=colors[idx],
                opacity=.6)
            data.append(trace)

        # Controls placement of y-axis tick labels
        tickvals = np.arange(len(data)) * spacing
        yaxis = dict(
            showgrid=False,
            zeroline=False,
            showline=False,
            showticklabels=True,
            tickmode='array',
            ticktext=columns,
            tickvals=tickvals,
            rangemode='nonnegative')

        xaxis = dict(
            showline=False,
            showgrid=False,
            showticklabels=True,
            linecolor='rgb(204, 204, 204)',
            autotick=False,
            ticks='outside',
            tickcolor='rgb(204, 204, 204)')

        if self.plot_mode == "div":
            layout = go.Layout(
                showlegend=False,
                height=max(42 * len(data), 600),
                xaxis=xaxis,
                yaxis=yaxis)
        else:
            layout = go.Layout(
                showlegend=False,
                height=max(42 * len(data), 600),
                xaxis=xaxis,
                yaxis=yaxis,
                title=title)

        # Reverse order since lastest plot is on the front
        fig = go.Figure(data=data[::-1], layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)


class CorrelationMatrix(MatrixPlotter):
    titlestring = "%s Correlation Matrix"
    shortname = "correlation"

    def plot(self, showticklabels=False):
        title = self.titlestring % (self.DS.name)
        D = self.DS.D.as_matrix().T
        xaxis = dict(
            title="Dimensions",
            ticks="",
            showgrid=False,
            zeroline=False,
            showticklabels=showticklabels,
        )
        yaxis = dict(
            scaleanchor="x",
            title="Dimensions",
            ticks="",
            showgrid=False,
            zeroline=False,
            showticklabels=showticklabels,
        )
        layout = dict(title=title, xaxis=xaxis, yaxis=yaxis)
        with np.errstate(divide='ignore', invalid='ignore'):
            C = np.nan_to_num(np.corrcoef(D))

        layout = self._get_layout(title, xaxis, yaxis)
        trace = go.Heatmap(
            x=self.DS.D.columns,
            y=self.DS.D.columns,
            z=C,
            zmin=-1,
            zmax=1,
            colorscale=self.BuRd)
        fig = dict(data=[trace], layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)


class ScreePlotter(MatrixPlotter):
    titlestring = "%s Scree Plot"
    shortname = "scree"

    def plot(self):
        title = self.titlestring % (self.DS.name)
        D = self.DS.D.as_matrix().T
        _, S, _ = np.linalg.svd(D, full_matrices=False)
        y = S
        x = np.arange(1, len(S) + 1)
        sy = np.sum(y)
        cy = np.cumsum(y)
        xaxis = dict(title='Factors')
        yaxis = dict(title='Proportion of Total Variance')
        var = go.Scatter(mode='lines+markers', x=x, y=y / sy, name="Variance")
        cumvar = go.Scatter(
            mode='lines+markers', x=x, y=cy / sy, name="Cumulative Variance")
        data = [var, cumvar]
        layout = self._get_layout(title, xaxis, yaxis)
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)


class HierarchicalClusterMeansDendrogram(MatrixPlotter):
    titlestring = "%s %s Cluster Means Dendrogram, Level %d"
    shortname = "cmd"

    def plot(self):
        title = self.titlestring % (self.DS.name, self.DS.clustname,
                                    self.DS.levels)
        self.shortname = self.DS.shortclustname + self.shortname
        means = []
        for c in self.DS.clusters[self.DS.levels]:
            means.append(np.average(c, axis=0))
        X = np.column_stack(means).T
        try:
            fig = ff.create_dendrogram(X)
        except:
            return '''
                <div class="row" style="margin-top:20%">
                    <div class="col-md-4 offset-md-4 text-center">
                        <h1><b>Only one cluster found.</b></h1>
                        <h3>Perhaps try another algorithm?</h2>
                </div>
                '''
        if self.plot_mode != "div":
            fig["layout"]["title"] = title

        fig["layout"]["xaxis"]["title"] = "Cluster Labels"
        fig["layout"]["yaxis"]["title"] = "Cluster Mean Distances"
        del fig.layout["width"]
        del fig.layout["height"]
        return self.makeplot(fig, "agg/" + self.shortname)


class HierarchicalStackedClusterMeansHeatmap(MatrixPlotter):
    titlestring = "%s %s Stacked Cluster Means, Level %d"
    shortname = "scmh"

    def plot(self, showticklabels=False):
        title = self.titlestring % (self.DS.name, self.DS.clustname,
                                    self.DS.levels)
        self.shortname = self.DS.shortclustname + self.shortname

        Xs = []
        for l in self.DS.clusters[1:self.DS.levels + 1]:
            #When number of samples is too high, need to downsample
            freq = [c.shape[0] for c in l]
            if sum(freq) > 500:
                freq = [round((x / sum(freq)) * 500) for x in freq]
                if sum(freq) != 500:  #Rounding can give numbers not exactly 500
                    freq[freq.index(max(freq))] += (500 - sum(freq))

            means = []
            for i, c in enumerate(l):
                means += [np.average(c, axis=0)] * freq[i]

            X = np.column_stack(means)
            Xs.append(X)
        X = np.vstack(Xs)[::-1, :]
        y_labels = np.tile(self.DS.columns,
                           X.shape[0] // len(self.DS.columns))[::-1]
        trace = go.Heatmap(
            z=X, zmin=-np.max(X), zmax=np.max(X), colorscale=self.BuRd)
        data = [trace]
        xaxis = go.XAxis(
            title="Clusters",
            showticklabels=False,
            ticks="",
            mirror=True,
            tickvals=[i for i in range(X.shape[1])])
        yaxis = go.YAxis(
            title="Dimensions",
            showticklabels=showticklabels,
            ticks="",
            ticktext=y_labels,
            tickvals=[i for i in range(len(y_labels))],
            mirror=True)
        emb_size = len(np.average(self.DS.clusters[0][0], axis=0))
        bar_locations = np.arange(0, X.shape[0] + emb_size - 1, emb_size) - 0.5
        shapes = [
            dict(type="line", x0=-0.5, x1=X.shape[1] - 0.5, y0=b, y1=b)
            for b in bar_locations
        ]
        if self.plot_mode == "div":
            layout = dict(xaxis=xaxis, yaxis=yaxis, shapes=shapes)
        else:
            layout = dict(title=title, xaxis=xaxis, yaxis=yaxis, shapes=shapes)
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)


class ClusterMeansLevelHeatmap(MatrixPlotter):
    titlestring = "%s %s Cluster Means, Level %d"
    shortname = "cmlh"

    def plot(self, showticklabels=False):
        title = self.titlestring % (self.DS.name, self.DS.clustname,
                                    self.DS.levels)
        self.shortname = self.DS.shortclustname + self.shortname

        #When number of samples is too high, need to downsample
        freq = [c.shape[0] for c in self.DS.clusters[self.DS.levels]]
        if sum(freq) > 500:
            freq = [round((x / sum(freq)) * 500) for x in freq]

        means = []
        for i, c in enumerate(self.DS.clusters[self.DS.levels]):
            means += [np.average(c, axis=0)] * freq[i]
        X = np.column_stack(means)
        trace = go.Heatmap(
            y=self.DS.columns[::-1],
            z=np.flipud(X),
            zmin=-np.max(X),
            zmax=np.max(X),
            colorscale=self.BuRd)
        data = [trace]
        xaxis = go.XAxis(
            title="Clusters",
            showticklabels=False,
            ticks="",
            mirror=True,
            tickvals=[i for i in range(X.shape[1])])
        yaxis = go.YAxis(
            title="Dimensions",
            showticklabels=showticklabels,
            ticks="",
            mirror=True)
        layout = self._get_layout(title=title, xaxis=xaxis, yaxis=yaxis)
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)


class ClusterMeansLevelLines(MatrixPlotter):
    titlestring = "%s %s Cluster Means, Level %d"
    shortname = "cmll"

    def plot(self, showticklabels=False):
        title = self.titlestring % (self.DS.name, self.DS.clustname,
                                    self.DS.levels)
        self.shortname = self.DS.shortclustname + self.shortname
        data = []
        colors = get_spaced_colors(len(self.DS.clusters[self.DS.levels]))

        #When number of samples is too high, need to downsample
        freq = [c.shape[0] for c in self.DS.clusters[self.DS.levels]]
        if sum(freq) > 300:
            freq = [round((x / sum(freq)) * 300) for x in freq]

        for i, c in enumerate(self.DS.clusters[self.DS.levels]):
            data.append(
                go.Scatter(
                    x=np.average(c, axis=0),
                    y=self.DS.columns,
                    mode="lines",
                    line=dict(width=np.sqrt(freq[i]), color=colors[i]),
                    name="cluster " + str(i)))
        xaxis = go.XAxis(
            title="Mean Values", showticklabels=False, mirror=True)
        yaxis = go.YAxis(
            title="Dimensions", showticklabels=showticklabels, mirror=True)
        layout = self._get_layout(title=title, xaxis=xaxis, yaxis=yaxis)
        fig = dict(data=data, layout=layout)
        return self.makeplot(fig, "agg/" + self.shortname)


class ClusterPairsPlot(MatrixPlotter):
    titlestring = "%s %s Classification Pairs Plot, Level %d"
    shortname = "cpp"

    def plot(self):
        title = self.titlestring % (self.DS.name, self.DS.clustname,
                                    self.DS.levels)
        self.shortname = self.DS.shortclustname + self.shortname
        data = []
        colors = get_spaced_colors(len(self.DS.clusters[self.DS.levels]))
        samples = []
        labels = []
        for i, c in enumerate(self.DS.clusters[self.DS.levels]):
            samples.append(c.T)
            labels.append(c.shape[0] * [i])
        samples = np.hstack(samples)[:3, :]
        labels = np.hstack(labels)
        df = pd.DataFrame(
            samples.T, columns=["Dim %d" % i for i in range(samples.shape[0])])
        df["label"] = ["Cluster %d" % i for i in labels]
        fig = ff.create_scatterplotmatrix(
            df, diag='box', index="label", colormap=colors)
        if self.plot_mode != "div":
            fig["layout"]["title"] = title
        else:
            fig["layout"]["title"] = None
        del fig.layout["width"]
        del fig.layout["height"]
        return self.makeplot(fig, "agg/" + self.shortname)



"""A setuptools based setup module.
See:
https://packaging.python.org/tutorials/distributing-packages/
https://github.com/pypa/sampleproject
"""

import sys

# Always prefer setuptools over distutils
from setuptools import setup

VERSION = '0.1.13.3'

# Install Cython
try:
    import Cython
except ImportError:
    #import pip
    # For installing Cython due to pip.main removal in Pip10
    import subprocess
    errno = subprocess.call([sys.executable, '-m', 'pip', 'install', 'cython'])
    if errno:
        print("Please install Cython.")
        raise SystemExit(errno)
    else:
        import Cython
        from Cython.Build import cythonize
        from Cython.Distutils import build_ext

setup(
    name='pymeda',
    version=VERSION,
    description='Matrix Exploratory Data Analysis',
    url='https://github.com/neurodata-nomads/pymeda',
    author='Jaewon Chung',
    author_email='j1c@jhu.edu',
    license='MIT',
    keywords='data visualization analysis',
    packages=['pymeda'],  # Required
    setup_requires=['Cython'],
    install_requires=[
        'Cython',
        'jinja2',
        'ipykernel',
        'pandas',
        'numpy',
        'scipy',
        'plotly==2.2.2',
        'matplotlib',
        'colorlover',
        'scikit-learn',
        'knor==0.0.1',
    ],
    package_data={
        'pymeda': ['*.html'],
    },
    include_package_data=True,
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Visualization',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
    ])
